% This code is licensed under the 3-clause BSD license.
% Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.
% See LICENSE.txt for details.

%%% GETTING READY ... %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

format long g; 
fprintf('AutoNetGen starts (seed %d)...\n\n',ridx); 
format short g;

std_A  = (max_A - min_A)/3;
mean_A = (max_A + min_A)/2;

nodes  = load('init.ang');
nodes  = [(1:size(nodes,1))',nodes,zeros(size(nodes,1),2),...
          ones(size(nodes,1),1)];
edges  = [];
layer  = 0;

finished = 0;

while ~finished

  nCurrent   = sum(nodes(:,4) == layer);
  nTotal     = size(nodes,1);
  nFormer    = nTotal - nCurrent;
  newEdges   = [floor(exprnd(newEdgesUni,nCurrent,1));floor(exprnd(...
                newEdgesBi,nCurrent*(nCurrent+1+(2*nFormer))/2,1))];
  count      = 0;

  %%% UNIMOLECULAR REACTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  for i = nFormer+1:nTotal

    m     = nodes(i,3);
    count = count + 1;
  
    for k = 1:newEdges(count)
      if size(edges,1) >= 2*maxEdges, finished = 1; continue; end
      % ISOMERIZATION
      if round(rand+0.5*isomerization) || m == 1 % a particle of mass m = 1
                                                 % is considered an atom and
                                                 % can, hence, not dissociate
                                                 % into two chemical species
        % create new vertex
        if round(rand+0.5*isomer_new) || ...
           sum(nodes(nodes(nodes(:,4) >= layer),3) == m) < 2
          A_node = nodes(i,5) + normrnd(mean_A,std_A);
          nodes  = [nodes;nodes(end,1)+1,0,m,layer+1,A_node,0];
  	  A_edge = max(nodes(i,5),A_node) + unifrnd(min_dA,max_dA);
          edges  = [edges;0,nodes(i,1),0,nodes(end,1),A_edge];
        % link to existing vertex
        else
          list   = nodes(nodes(:,3) == m,:);
       	  list   = list(list(:,4) >= layer);
          list   = list(list ~= i);
  	  idx    = nodes(list(randi([1,length(list)])));
          dA     = abs(nodes(i,5) - nodes(idx,5));
  	  if 3 * abs(normrnd(0,std_A)) > dA
  	    A_edge = max(nodes(i,5),nodes(idx,5)) + unifrnd(min_dA,max_dA);
  	    edges  = [edges;0,nodes(i,1),0,idx,A_edge];
  	  end
        end
      % DISSOCIATION
      else
        M = randi([1 floor(m/2)]);
        % create two new vertices
        if round(rand+0.5*diss_2new) ...
           || ~sum(nodes(nodes(nodes(:,4) >= layer),3) == M) ...
          A_nodes = nodes(i,5) + normrnd(mean_A,std_A/sqrt(2));
  	  A_node1 = A_nodes + normrnd(0,std_A/sqrt(2));
  	  A_node2 = A_nodes - A_node1; 
          nodes   = [nodes;nodes(end,1)+1,0,M  ,layer+1,A_node1,0];
          nodes   = [nodes;nodes(end,1)+1,0,m-M,layer+1,A_node2,0];
          A_edge  = max(nodes(i,5),A_nodes) + unifrnd(min_dA,max_dA);
          edges   = [edges;0,nodes(i,1),nodes(end-1,1),nodes(end,1),...
                     A_edge];
        % create one new vertex and link to one existing vertex
        elseif round(rand+0.5*diss_1new) || ~sum(nodes(:,3) == m-M) 
          list    = nodes(nodes(:,3) == M,:);
  	  list    = list(list(:,4) >= layer);
          idx     = nodes(list(randi([1,length(list)])));
  	  A_nodes = nodes(i,5) + normrnd(mean_A,std_A);
  	  A_node  = A_nodes - nodes(idx,5);
          nodes   = [nodes;nodes(end,1)+1,0,m-M,layer+1,A_node,0];
  	  A_edge  = max(nodes(i,5),A_nodes) + unifrnd(min_dA,max_dA); 
          edges   = [edges;0,nodes(i,1),idx,nodes(end,1),A_edge];
        % link to two existing vertices
        else
          list1   = nodes(nodes(:,3) == M,:);
  	  list1   = list1(list1(:,4) >= layer);
  	  if M == m-M
  	    list2 = list1;
  	  else
  	    list2 = nodes(nodes(:,3) == m-M);
  	  end
	  idx1  = nodes(list1(randi([1,length(list1)])));
  	  idx2  = nodes(list2(randi([1,length(list2)])));
  	  dA    = abs(nodes(i,5) - (nodes(idx1,5) + nodes(idx2,5)));
  	  if 3 * abs(normrnd(0,std_A)) > dA
  	    A_edge = max(nodes(i,5),nodes(idx1,5)+nodes(idx2,5)) ...
  	           + unifrnd(min_dA,max_dA);
  	    edges  = [edges;0,nodes(i,1),min(idx1,idx2),max(idx1,idx2),...
                      A_edge];
  	  end
        end
      end
    end
  
  end

  %%% BIMOLECULAR REACTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  for i = 1:nTotal, for j = nFormer+1:nTotal, if j >= i

    m     = nodes(i,3) + nodes(j,3);
    count = count + 1;
  
    for k = 1:newEdges(count)
      if size(edges,1) >= 2*maxEdges, finished = 1; continue; end
      % ASSOCIATION
      if round(rand+0.5*association) || m == 2 % m = 2 implies two particles
                                               % of mass m = 1, which are
                                               % considered atoms and, hence,
                                               % cannot undergo substitution
        % create new vertex
        if round(rand+0.5*ass_new) || ...
           ~sum(nodes(nodes(nodes(:,4) >= layer),3) == m)
          A_node = nodes(i,5) + nodes(j,5) + normrnd(mean_A,std_A);
          nodes  = [nodes;nodes(end,1)+1,0,m,layer+1,A_node,0];
          A_edge = max(nodes(i,5)+nodes(j,5),A_node) ...
                 + unifrnd(min_dA,max_dA);
          edges  = [edges;nodes(i,1),nodes(j,1),0,nodes(end,1),A_edge];
        % link to existing vertex
        else
          list   = nodes(nodes(:,3) == m,:);
          list   = list(list(:,4) >= layer);
          idx    = nodes(list(randi([1,length(list)])));
  	  dA     = abs(nodes(i,5) + nodes(j,5) - nodes(idx,5));
  	  if 3 * abs(normrnd(0,std_A)) > dA
  	    A_edge = max(nodes(i,5)+nodes(j,5),nodes(idx,5)) ...
                   + unifrnd(min_dA,max_dA);
  	    edges  = [edges;nodes(i,1),nodes(j,1),0,idx,A_edge];
          end
        end
      % (SINGLE) SUBSTITUTION
      else
        if nodes(i,3) == nodes(j,3)
          M = randi([1 floor((m-1)/2)]);
          % we chose "m-1" because a single substitution between two species
  	  % with the same mass cannot produce two products with the same mass
  	  % (examples: protonation, SN2 reaction)
        else
          M = randi([1 floor(m/2)]);
        end
        % create two new vertices
        if round(rand+0.5*disp_2new) ...
           || sum(nodes(nodes(nodes(:,4) >= layer),3) == M) < 2
          A_nodes = nodes(i,5) + nodes(j,5) + normrnd(mean_A,std_A/sqrt(2));
       	  A_node1 = A_nodes + normrnd(0,std_A/sqrt(2));
  	  A_node2 = A_nodes - A_node1; 
          nodes   = [nodes;nodes(end,1)+1,0,M  ,layer+1,A_node1,0];
          nodes   = [nodes;nodes(end,1)+1,0,m-M,layer+1,A_node2,0];
  	  A_edge  = max(nodes(i,5)+nodes(j,5),A_nodes) ...
                  + unifrnd(min_dA,max_dA);
          edges   = [edges;nodes(i,1),nodes(j,1),...
  	             nodes(end-1,1),nodes(end,1),A_edge];
        % create one new vertex and link to one existing vertex
        elseif round(rand+0.5*disp_1new) || sum(nodes(:,3) == m-M) < 2
          list    = nodes(nodes(:,3) == M,:);
  	  list    = list(list(:,4) >= layer);
  	  list    = list(list ~= i);
  	  list    = list(list ~= j);
          idx     = nodes(list(randi([1,length(list)])));
          A_nodes = nodes(i,5) + nodes(j,5) + normrnd(mean_A,std_A);
  	  A_node  = A_nodes - nodes(idx,5);
          nodes   = [nodes;nodes(end,1)+1,0,m-M,layer+1,A_node,0];
  	  A_edge  = max(nodes(i,5)+nodes(j,5),A_nodes) ...
                  + unifrnd(min_dA,max_dA); 
          edges   = [edges;nodes(i,1),nodes(j,1),idx,nodes(end,1),...
                     A_edge];
        % link to two existing vertices
        else
          list1   = nodes(nodes(:,3) == M  ,:);
  	  list1   = list1(list1(:,4) >= layer);
  	  list1   = list1(list1 ~= i);
  	  list1   = list1(list1 ~= j);
  	  if M == m-M
  	    list2 = list1;
  	  else
  	    list2 = nodes(nodes(:,3) == m-M);
  	    list2 = list2(list2 ~= i);
  	    list2 = list2(list2 ~= j);
  	  end
  	  idx1  = nodes(list1(randi([1,length(list1)])));
  	  idx2  = nodes(list2(randi([1,length(list2)])));
  	  dA    = abs(nodes(i,5) + nodes(j,5) ...
                - (nodes(idx1,5) + nodes(idx2,5)));
  	  if 3 * abs(normrnd(0,std_A)) > dA
  	    A_edge = max(nodes(i,5)+nodes(j,5),nodes(idx1,5)+nodes(idx2,5)) ...
  	           + unifrnd(min_dA,max_dA);
  	    edges  = [edges;nodes(i,1),nodes(j,1),...
  	              min(idx1,idx2),max(idx1,idx2),A_edge];
  	  end
        end
      end
    end
  
  end, end, end
  
  %%% NETWORK PROCESSING %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  index = [];
  
  for i = 1:size(edges,1)
    if ~(edges(i,1) + edges(i,3))
      if edges(i,2) == edges(i,4)
        index = [index;i];
      end
    elseif (edges(i,1) == edges(i,3)) || (edges(i,1) == edges(i,4)) || ...
           (edges(i,2) == edges(i,3)) || (edges(i,2) == edges(i,4))
      index = [index;i];
    end
  end
  
  index = sort(index,'descend');
  
  for i = 1:length(index)
    edges = [edges(1:index(i)-1,:);edges(index(i)+1:end,:)];
  end
  
  index = [];

  for l = 1:size(edges,1)
    for m = l+1:size(edges,1)
      if (edges(l,1) == edges(m,1) && edges(l,2) == edges(m,2) && ...
          edges(l,3) == edges(m,3) && edges(l,4) == edges(m,4)) || ...
         (edges(l,3) == edges(m,1) && edges(l,4) == edges(m,2) && ...
          edges(l,1) == edges(m,3) && edges(l,2) == edges(m,4))
        index = [index;m];
      end
    end
  end

  index = unique(index);
  tmp   = [];
  
  for l = 1:size(edges,1)
    if ~sum(l == index)
      tmp = [tmp;edges(l,:)];
    end
  end

  edges = tmp;
  
  if size(nodes,1) > nTotal
    layer = layer + 1;
  end

end

edges = edges(1:maxEdges,:); % needed for "meta.m" regeneration case

%%% SAMPLING OF ACTIVATION ENERGIES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

A0_node = nodes(:,5);
A0_edge = edges(:,5);
A0      = abs(actEnergy(A0_node,A0_edge,edges(:,1:4)));
k0      = rateConstant(A0,T);

covA    = covmat(meanErr_A,size(nodes,1));
Ab_node = mvnrnd(nodes(:,5),covA,B)';
Ab_edge = edgeEnergy(nodes(:,5),Ab_node,edges(:,5),edges(:,1:4),meanErr_A);
Ab      = abs(actEnergy(Ab_node,Ab_edge,edges(:,1:4)));
kb      = rateConstant(Ab,T);

k       = [k0,kb];
A_node  = [A0_node,Ab_node];
A_edge  = [A0_edge,Ab_edge];
dA      = [A0,Ab];

%%% ALMOST DONE ... %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nodes_crn = nodes(:,[1:3,6]);
edges_crn = [edges(:,1:4),k(1:size(edges,1),:),k(size(edges,1)+1:end,:)];

save -ascii crn/nodes.crn nodes_crn
save -ascii crn/edges.crn edges_crn
save -ascii crn/A_node.crn A_node
save -ascii crn/A_edge.crn A_edge
