% This code is licensed under the 3-clause BSD license.
% Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.
% See LICENSE.txt for details.

function J = jacobi(t,y,rxns,k)

  N  = size(y,1);
  y  = [1;y];
  L  = size(rxns,1);
  r  = rxns + 1;
  S  = stoichMatrix(r);
  df = zeros(L,N);
  y  = y(r);

  if L == 1
    y = y';
  end

  for l = 1:L
    for n = 2:N+1
      if ~isempty(find(r(l,:) == n))
        df(l,n-1) = ((r(l,1) == n) + (r(l,2) == n)) * k(l,1) ...
                  * y(l,1) ^  (r(l,1) ~= n) ...
                  * y(l,2) ^ ((r(l,2) ~= n) || (r(l,1) == r(l,2))) ...
                  - ((r(l,3) == n) + (r(l,4) == n)) * k(L+l,1) ...
                  * y(l,3) ^  (r(l,3) ~= n) ...
                  * y(l,4) ^ ((r(l,4) ~= n) || (r(l,3) == r(l,4)));
      end
    end
  end

  J = S * df;

end
