% This code is licensed under the 3-clause BSD license.
% Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.
% See LICENSE.txt for details.

clean

%%% AutoNetGen (generation of complete network) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%param_ang
%autonetgen

%results = [ridx,mean(std(Ab,0,2)),std(std(Ab,0,2))]; % meta

%clear_autonetgen

%%% KiNetX (applied to complete network) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

param_knx
prepare

f_analys = 0; % default: 0
s_analys = 0; % default: 0

kinetx

%s_analys = 0; % delete

%results = [results,N,L]; % meta

N_full    = N;
L_full    = L;
rxns_full = rxns;
k_full    = k;
spec_full = spec;
ys_full   = ys;

plot_knx

%%% KiNetX-guided network exploration %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fprintf('KiNetX-guided network exploration starts...\n\n');

finished = 0; % default: 0
attempt  = 0; % default: 0

%%% Initial reaction network is being prepared...

l_idx_exp = [];

for l = 1:L_full
  for n = spec(~~spec(:,4))'
    if sum(n == edges(l,1:4)+1)
      if prod(nodes(edges(l,1:2)+1,4)) || prod(nodes(edges(l,1:2)+1,4))
        l_idx_exp = [l_idx_exp;l];
      end
    end
  end
end

l_idx_exp = unique(l_idx_exp);

%%% Iterative exploration starts...

while ~finished

  clear_kinetx

  attempt = attempt + 1;

  fprintf('Exploration step no. %d:\n\n',attempt)

  explored = 0;

  rxns = rxns_full(l_idx_exp,:);
  k    = [k_full(l_idx_exp,:);k_full(L_full+l_idx_exp,:)];
  spec = nodes(2:max(rxns(:))+1,:);
  N    = length(unique(rxns)) - sum(~unique(rxns));
  L    = size(rxns,1);

  kinetx
%  plot_knx

  for b = 1:B
    delta_exp(b,1) = metric_u(ys_full(:,:,b),ys(:,:,b),U);
    Delta_exp(b,1) = metric(ys_full(:,:,b),ys(:,:,b),dt,t_max);
  end

  dDelta_exp       = max(delta_exp,Delta_exp);
  dDelta_exp_quant = quantile(dDelta_exp,(1+clevel)/2);

  if dDelta_exp_quant < y_min
    fprintf('Good news: Network has been successfully explored!\n\n')
    fprintf('Final KiNetX analysis starts...\n\n');
    if ~exist('max_g_max')
      max_g_max = NaN;
    end
    %results = [results,max_g_max]; % meta
    explored = 1;
    finished = 1;
    f_analys = 1;
    s_analys = 1;
    kinetx
    plot_knx
  else
    max_g_max = max(g_max(:)); % meta
  end

  count = 1;

  while ~explored

    nodes(g_max_ranking(count)+1,4) = 1;

    l_idx_exp = [];

    for l = 1:L_full
      for n = (find(nodes(2:end,4))+1)'
        if sum(n == edges(l,1:4)+1)
          if prod(nodes(edges(l,1:2)+1,4)) || prod(nodes(edges(l,3:4)+1,4))
            l_idx_exp = [l_idx_exp;l];
          end
        end
      end
    end

    l_idx_exp = unique(l_idx_exp);

    if sum(l_idx_exp) > L
      explored = 1;
    elseif count < length(g_max_ranking) && ~nodes(g_max_ranking(count+1)+1,4)
      count = count + 1;
    else
      explored = 1;
      finished = 1;
    end

  end

end

results = [results,N,L,N_,L_];

%%% Sensitivity analysis on final network + subsequent KiNetX run...

if s_analys

%  l_idx_exp = (1:L_full)'; % delete

  if exist('l_idx_dfa')
    dfa_needed = 1;
  else
    dfa_needed       = 0;
    dDelta_dfa_quant = 0;
  end
  
  %meta
  l_crit_0 = l_idx_exp(l_idx_xfa(l_idx_msa_0));
  
  l_crit      = l_idx_exp(l_idx_xfa(l_idx_msa));
  n_crit(:,1) = unique(edges(l_crit,1:4)+1); 

  X  = zeros(N_full,1); X(l_crit)  = 1;
  X0 = zeros(N_full,1); X0(l_crit_0) = 1;

  results = [results,attempt,deltaB,DeltaB,min(Y),max(Y),dfa_needed,...
             dDelta_cfa_quant,dDelta_dfa_quant,min(sum(G > G_min)),...
             max(sum(G > G_min)),length(l_idx_msa),sum(abs(X - X0))]; % meta

  f_analys = 0;
  s_analys = 0;

%  B = 6; % delete

  A_node  = [zeros(1,B);load('crn/A_node.crn')];
  A_edge  = load('crn/A_edge.crn');

  A_edge(l_crit,:) = repmat(mean(A_edge(l_crit,:),2),1,B);
  A_node(n_crit,:) = repmat(mean(A_node(n_crit,:),2),1,B);

  A_edge_exact = mean(A_edge,2);
  A_node_exact = mean(A_node,2);

  dA = abs(actEnergy(A_node(2:end,:),A_edge,edges(:,1:4)));
  k  = rateConstant(dA,load('crn/temp.crn'));
  k  = [k(l_idx_exp(l_idx_xfa),:);k(L_full+l_idx_exp(l_idx_xfa),:)];

  dA_exact = abs(actEnergy(A_node_exact(2:end),A_edge_exact,edges(:,1:4)));
  k_exact  = rateConstant(dA_exact,load('crn/temp.crn'));
  ys_exact = solver(rxns_full,k_exact,y0,tspan,ts,tol);

  rxns = rxns_;
  spec = spec_;
  N    = N_;
  L    = L_;

  clear_kinetx
  clear n_crit

  fprintf('KiNetX detected %d critical reactions.\n',length(l_crit))
  fprintf('Reevaluation of critical reactions...\n\n')
  fprintf('KiNetX analysis on refined model starts...\n\n')

%  B = 1; % delete

  kinetx
  plot_knx

  deltaX = metric_u(mean(ys(:,:,:,1),3),ys_exact,U);
  DeltaX = metric(mean(ys(:,:,:,1),3),ys_exact,dt,t_max);

  results = [results,deltaB,DeltaB,deltaX,DeltaX]; % meta
end
