% This code is licensed under the 3-clause BSD license.
% Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.
% See LICENSE.txt for details.

function ys = solver(rxns,k,y0,tspan,ts,tol)

  L = size(rxns,1);
  N = length(y0);

  idx = unique([find(y0 > 0);unique(rxns(:))]); 
  if sum(~idx), idx = idx(2:end); end
  y0  = y0(idx);

  for l = 1:L
    for i = 1:4
      if rxns(l,i)
        rxns(l,i) = find(rxns(l,i) == idx);
      end
    end
  end

  %opts  = odeset('Jacobian',@(t,y) jacobi(t,y,rxns,k),'RelTol',tol.rel,...
  %               'AbsTol',tol.abs,'NonNegative',(1:length(y0)));
  opts  = odeset('RelTol',tol.rel,'AbsTol',tol.abs,'NonNegative',...
                 (1:length(y0)));
  [t y] = ode15s(@(t,y) odefcn(t,y,rxns,k),tspan,y0,opts);

  U  = length(ts);
  ys = zeros(U,N);

  for n = 1:N
    if sum(n == idx)
      ys(:,n) = spline(t,y(:,find(n == idx)),ts);
    end
  end
 
end
